"""UEVaultManager"""

__name__ = 'UEVaultManager'
__version__ = '1.0.1'
__codename__ = 'Dragon+'
__author__ = 'Laurent Ongaro'
__author_email__ = 'laurent@gameamea.com'
__description__ = 'Free and open-source replacement for the Epic Games Launcher application, mainly to manage the assets for Unreal Engine'
__copyright__ = '2023 Laurent Ongaro'
__license__ = 'GPL-3'
__url__ = 'https://github.com/LaurentOngaro/UEVaultManager'
